local Maze = {}

local function StartMaze( Width, Height, StartX, StartY )
	StartX = StartX or love.math.random( Width )
	StartY = StartY or love.math.random( Height )
	
	local Grid = {}
	for y = 1, Height do
		Grid[y] = {}
		for x = 1, Width do
			Grid[y][x] = {
				LeftWall = true, 
				RightWall = true, 
				TopWall = true, 
				BottomWall = true, 
				Visited = false, 
			}
		end
	end
	local VisitedCells = {}
	
	return StartX, StartY, Grid, VisitedCells
end

function Maze.New( Width, Height, StartX, StartY ) 
	StartX, StartY, Grid, VisitedCells = StartMaze( Width, Height, StartX, StartY )
	
	Maze.Generate( StartX, StartY, Grid, VisitedCells, 1 )
	
	return Grid
end

function Maze.NewStep( Width, Height, StartX, StartY ) 
	StartX, StartY, Grid, VisitedCells = StartMaze( Width, Height, StartX, StartY )
	
	local x, y, ReverseAmount = Maze.Step( StartX, StartY, Grid, VisitedCells, 1 )
	
	return x, y, Grid, VisitedCells, ReverseAmount
end

local function GatherInformation( x, y, Grid, VisitedCells, ReverseAmount )
	Grid[y][x].Visited = true
	local NeighborCells = Maze.GetNeighborCells( x, y, Grid )
	local PossibleOptions = Maze.ProcessNeighborCells( NeighborCells )
	local Cell, ReverseAmount = Maze.ChooseCell( PossibleOptions, VisitedCells, Grid, ReverseAmount )
	
	if not Cell then return end
	
	if not Maze.UnvisitedCellsRemain( Grid ) then return end
	
	Maze.RemoveCorrectWall( Cell.Direction, x, y, Grid )
	Maze.RemoveCorrectWall( Maze.OppositeWall( Cell.Direction ), Cell.x, Cell.y, Grid )
	
	VisitedCells[ #VisitedCells + 1 ] = { x = x, y = y, Cell = Cell }
	--[[
	print( x, y, Cell.Direction, Cell.x, Cell.y )
	print( 'LRUD: Current: ', Grid[y][x].LeftWall, Grid[y][x].RightWall, Grid[y][x].TopWall, Grid[y][x].BottomWall )
	print( 'LRUD: Next: ', Grid[Cell.y][Cell.x].LeftWall, Grid[Cell.y][Cell.x].RightWall, Grid[Cell.y][Cell.x].TopWall, Grid[Cell.y][Cell.x].BottomWall )
	print( '--' )
	]]
	
	return Cell, ReverseAmount
end

function Maze.Step( x, y, Grid, VisitedCells, ReverseAmount )
	local Cell, ReverseAmount = GatherInformation( x, y, Grid, VisitedCells, ReverseAmount )
	if not Cell then return false end
	return Cell.x, Cell.y, ReverseAmount
end

function Maze.Generate( x, y, Grid, VisitedCells, ReverseAmount )
	local Cell, ReverseAmount = GatherInformation( x, y, Grid, VisitedCells, ReverseAmount )
	Maze.Generate( Cell.x, Cell.y, Grid, VisitedCells )
end

function Maze.GetNeighborCells( x, y, Grid ) 
	return {
		{ Condition = Grid[y][x - 1] or nil, x = x - 1, y = y }, -- Left
		{ Condition = Grid[y][x + 1] or nil, x = x + 1, y = y }, -- Right
		{ Condition = Grid[y - 1] and Grid[y - 1][x] or nil, x = x, y = y - 1 }, -- Up
		{ Condition = Grid[y + 1] and Grid[y + 1][x] or nil, x = x, y = y + 1 }, -- Down
	}
end

function Maze.ProcessNeighborCells( NeighborCells ) 
	local PossibleOptions = {}
	for Index = 1, #NeighborCells do
		Cell = NeighborCells[Index]
		if Cell.Condition and not Cell.Condition.Visited then
			local Direction
			if Index == 1 then 
				Direction = 'Left'
			elseif Index == 2 then
				Direction = 'Right'
			elseif Index == 3 then
				Direction = 'Up'
			elseif Index == 4 then
				Direction = 'Down'
			end
			
			PossibleOptions[ #PossibleOptions + 1 ] = { x = Cell.x, y = Cell.y, Direction = Direction }
		end
	end
	return PossibleOptions
end

function Maze.ChooseCell( Options, VisitedCells, Grid, ReverseAmount )
	if #Options < 1 then
		local Val = Maze.GetNearestViableOption( VisitedCells, Grid )
		if Val then
			Val.Direction = nil
			return Val, 1
		else
			local Val = VisitedCells[ #VisitedCells - ReverseAmount ]
			if not Val then return end
			Val.Direction = nil
			return Val, ReverseAmount + 2
		end
	end
	return Options[ love.math.random( #Options ) ], 1
end

function Maze.GetNearestViableOption( VisitedCells, Grid )
	for a = #VisitedCells, 1, -1 do
		local Visited = VisitedCells[a]
		local Choices = Maze.ProcessNeighborCells( Maze.GetNeighborCells( Visited.x, Visited.y, Grid ) )
		if #Choices > 0 then return Visited end
	end
end

function Maze.UnvisitedCellsRemain( Grid )
	for y = 1, #Grid do
		for x = 1, #Grid[y] do
			if not Grid[y][x].Visited then
				return true
			end
		end
	end
	return false
end

function Maze.RemoveCorrectWall( Direction, x, y, Grid )
	if Direction == 'Left' then
		Maze.RemoveLeftWall( x, y, Grid )
	elseif Direction == 'Right' then
		Maze.RemoveRightWall( x, y, Grid )
	elseif Direction == 'Up' then
		Maze.RemoveTopWall( x, y, Grid )
	elseif Direction == 'Down' then
		Maze.RemoveBottomWall( x, y, Grid )
	end
end

function Maze.OppositeWall( Direction )
	if Direction == 'Left' then
		return 'Right'
	elseif Direction == 'Right' then
		return 'Left'
	elseif Direction == 'Up' then
		return 'Down'
	elseif Direction == 'Down' then
		return 'Up'
	end
end

function Maze.RemoveLeftWall( x, y, Grid )
	Grid[y][x].LeftWall = false
end

function Maze.RemoveRightWall( x, y, Grid )
	Grid[y][x].RightWall = false
end

function Maze.RemoveTopWall( x, y, Grid )
	Grid[y][x].TopWall = false
end

function Maze.RemoveBottomWall( x, y, Grid )
	Grid[y][x].BottomWall = false
end

return Maze